@@ -165,14 +165,14 @@ stop_update_timer(CCCHHDoorIsDesktopWidget *self)
 
 // install libconic connection event handler
 // returns TRUE if handler was installed
-// returns FALSE if handler was already set
+// returns FALSE if handler was already set or !self->connection
 // TODO: check return code from g_signal_connect
 static inline gboolean
 install_connection_handler(CCCHHDoorIsDesktopWidget *self)
 {
 	g_assert(IS_CCCHH_DOORIS_DESKTOP_WIDGET(self));
 
-	if (!self->connection_handler) {
+	if (!self->connection_handler && self->connection) {
 		// set connected to FALSE, initial event may set it to TRUE
 		self->connected = FALSE;
 		// Connect signal to receive connection events
@@ -190,13 +190,13 @@ install_connection_handler(CCCHHDoorIsDesktopWidget *self)
 
 // remove libconic connection event handler
 // returns TRUE if handler was removed
-// returns FALSE if no handler was set
+// returns FALSE if no handler was set or !self->connection
 static inline gboolean
 remove_connection_handler(CCCHHDoorIsDesktopWidget *self)
 {
 	g_assert(IS_CCCHH_DOORIS_DESKTOP_WIDGET(self));
 
-	if (self->connection_handler) {
+	if (self->connection_handler && self->connection) {
 		g_object_set(self->connection, "automatic-connection-events", FALSE, NULL);
 		g_signal_handler_disconnect(self->connection, self->connection_handler);
 		self->connection_handler = 0;
@@ -240,7 +240,7 @@ load_settings(CCCHHDoorIsDesktopWidget *self)
 		}
 
 		self->update_url = gnome_vfs_make_uri_from_input(strval);
-		g_debug(WIDGET_NAME": load_settings: url = %s", self->update_url);
+		g_debug(WIDGET_NAME": load_settings: update_url = %s", self->update_url);
 
 		g_free(strval);
 		strval = NULL;
@@ -263,7 +263,7 @@ load_settings(CCCHHDoorIsDesktopWidget *self)
 			g_warning(WIDGET_NAME": self->update_interval > UPDATE_INTERVAL_MAX, maybe a bug?");
 			self->update_interval = UPDATE_INTERVAL_MAX;
 		}
-		g_debug(WIDGET_NAME": load_settings: update = %ld", (long)self->update_interval);
+		g_debug(WIDGET_NAME": load_settings: update_interval = %ld", (long)self->update_interval);
 	}
 
 	gboolean boolval = gconf_client_get_bool(self->gconf_client,
@@ -276,7 +276,7 @@ load_settings(CCCHHDoorIsDesktopWidget *self)
 		error = NULL;
 	} else {
 		self->notify_enabled = boolval;
-		g_debug(WIDGET_NAME": load_settings: notify = %d", self->notify_enabled);
+		g_debug(WIDGET_NAME": load_settings: notify_enabled = %d", self->notify_enabled);
 	}
 
 
@@ -414,7 +414,6 @@ save_settings(CCCHHDoorIsDesktopWidget *self)
 }
 
 
-// TODO: skip draw if not on active desktop
 static void
 draw (CCCHHDoorIsDesktopWidget *self, cairo_t *cr)
 {
@@ -422,7 +421,6 @@ draw (CCCHHDoorIsDesktopWidget *self, cairo_t *cr)
 
 	g_debug(WIDGET_NAME": draw()");
 
-	// TODO: handled in expose event and force_redraw() already
 	if (!self->desktop_visible) {
 		g_debug(WIDGET_NAME": desktop not visible, not drawing");
 		return;
@@ -430,6 +428,10 @@ draw (CCCHHDoorIsDesktopWidget *self, cairo_t *cr)
 
 	if (!self->image_current) {
 		g_debug(WIDGET_NAME": draw() without image");
+		if (self->updating) {
+			g_debug(WIDGET_NAME": update in progress, cancelling draw");
+			return;
+		}
 		self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_FNORD);
 	}
 
@@ -584,6 +586,8 @@ fetch_update_lock (CCCHHDoorIsDesktopWidget *self)
 		return FALSE;
 	}
 
+	self->updating = TRUE;
+
 	// indicate update in progress, make icon transparent
 	self->current_alpha = IMAGE_ALPHA_PROGRESS;
 	force_redraw(self);
@@ -602,6 +606,8 @@ fetch_update_unlock (CCCHHDoorIsDesktopWidget *self)
 	self->current_alpha = IMAGE_ALPHA_NORMAL;
 	force_redraw(self);
 
+	self->updating = FALSE;
+
 	g_static_mutex_unlock(&(self->fetch_lock));
 }
 
@@ -650,6 +656,7 @@ fetch_update_cancel_cb (gpointer data)
   
 	g_debug(WIDGET_NAME": fetch_update_cancel_cb()");
 
+	// TODO: explain this
 	if (g_source_is_destroyed(g_main_current_source())) {
 		// this shouldn't happen
 		g_warning(WIDGET_NAME": update_cancel: source destroyed");
@@ -676,6 +683,7 @@ fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 	g_debug(WIDGET_NAME": fetch_update_async_open_cb()");
 
+	// stop the timeout event
 	if (self->gvfs_timeout_handler) {
 		g_source_remove (self->gvfs_timeout_handler);
 		self->gvfs_timeout_handler = 0;
@@ -687,7 +695,7 @@ fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 		return;
 	}
 
-	// reuse input buffer, ensure it has at least MAX_HTTP_SIZE+1
+	// reuse input buffer, ensure it holds at least MAX_HTTP_SIZE+1 bytes
 	if (!(self->gvfs_input_buffer = g_try_realloc(self->gvfs_input_buffer, MAX_HTTP_SIZE+1))) {
 		g_warning(WIDGET_NAME": failed to allocate input buffer");
 		fetch_update_process(self, DOORIS_FNORD, 0L, 0L);
@@ -698,6 +706,8 @@ fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 	self->gvfs_timeout_handler = g_timeout_add_seconds(TIMEOUT_READ, fetch_update_cancel_cb, self);
 
+	// read data
+	// next callback: either fetch_update_async_read_cb() or fetch_update_cancel_cb()
 	gnome_vfs_async_read(handle, self->gvfs_input_buffer, MAX_HTTP_SIZE,
 		fetch_update_async_read_cb, self);
 }
@@ -811,7 +821,7 @@ fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 // handle async close callback
 static void
-fetch_update_async_close_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
+fetch_update_async_close_cb (GnomeVFSAsyncHandle *handle G_GNUC_UNUSED, GnomeVFSResult result,
 	gpointer data)
 {
 	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (data);
@@ -866,7 +876,7 @@ fetch_update_process (CCCHHDoorIsDesktopWidget *self, dooris_state_t new_state,
 		return;
 	}
 
-	// state has changed since last update
+	// state changed since last update
 
 	// release old icon surface
 	if (self->image_current) {
@@ -929,13 +939,14 @@ ccchh_dooris_desktop_widget_on_timeout (gpointer data)
   
 	g_debug(WIDGET_NAME": ccchh_dooris_desktop_widget_on_timeout()");
 
+	// TODO: explain this
 	if (g_source_is_destroyed(g_main_current_source())) {
 		// this shouldn't happen
 		g_warning(WIDGET_NAME": on_timeout: source destroyed");
 		return FALSE;
 	}
 
-	// stop timer when offline
+	// stop timer if offline
 	// connection event starts it again
 	if (!self->connected) {
 		g_debug(WIDGET_NAME": not connected, stopping update timer");
@@ -962,11 +973,22 @@ ccchh_dooris_desktop_widget_expose_event (GtkWidget *widget, GdkEventExpose *eve
 
 	if (!self->desktop_visible) {
 		g_debug(WIDGET_NAME": desktop not visible, ignoring expose event");
-		return;
+		return FALSE;
+	}
+
+	// TODO: make sure we don't miss a necessary redraw
+	if (!g_static_mutex_trylock(&(self->expose_lock))) {
+		g_debug(WIDGET_NAME": expose_lock set, ignoring event");
+		return FALSE;
 	}
 
 	if (!self->image_current) {
 		g_debug(WIDGET_NAME": expose event without image");
+		if (self->updating) {
+			g_debug(WIDGET_NAME": update in progress, cancelling expose event");
+			g_static_mutex_unlock(&(self->expose_lock));
+			return FALSE;
+		}
 		self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_FNORD);
 		// no state yet, begin update if enabled
 		if (self->auto_update)
@@ -988,6 +1010,10 @@ ccchh_dooris_desktop_widget_expose_event (GtkWidget *widget, GdkEventExpose *eve
 	cairo_destroy (cr);
 
 	self->redraw_required = FALSE;
+
+	g_debug(WIDGET_NAME": expose event finished");
+	g_static_mutex_unlock(&(self->expose_lock));
+
 	return FALSE;
 }
 
@@ -995,7 +1021,7 @@ ccchh_dooris_desktop_widget_expose_event (GtkWidget *widget, GdkEventExpose *eve
 // button press event - highlight icon
 static gboolean 
 ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget, 
-  GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
+  GdkEventButton *event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
 {
 	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (widget);
 
@@ -1003,7 +1029,7 @@ ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget,
  
 	// TODO: do we really need a button lock? fetch_update locks itself
 	if (!g_static_mutex_trylock(&(self->button_lock))) {
-		g_debug(WIDGET_NAME": button_lock set, not updating");
+		g_debug(WIDGET_NAME": button_lock set, ignoring event");
 		return TRUE;
 	}
 
@@ -1019,7 +1045,7 @@ ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget,
 //       i.e. no previous button-press
 static gboolean 
 ccchh_dooris_desktop_widget_on_button_leave_event (GtkWidget *widget, 
-  GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
+  GdkEventButton *event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
 {
 	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (widget);
 
@@ -1037,7 +1063,7 @@ ccchh_dooris_desktop_widget_on_button_leave_event (GtkWidget *widget,
 // button release event - begin update, remove highlight
 static gboolean 
 ccchh_dooris_desktop_widget_on_button_release_event (GtkWidget *widget, 
-  GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
+  GdkEventButton *event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
 {
 	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (widget);
 
@@ -1077,7 +1103,7 @@ ccchh_dooris_desktop_widget_on_button_release_event (GtkWidget *widget,
 
 // libconic connection event
 static void
-ccchh_dooris_desktop_widget_connection_event(ConIcConnection *connection,
+ccchh_dooris_desktop_widget_connection_event(ConIcConnection *connection G_GNUC_UNUSED,
   ConIcConnectionEvent *event, gpointer user_data)
 {
 	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (user_data);
@@ -1340,7 +1366,7 @@ ccchh_dooris_desktop_widget_on_settings_dialog_response (GtkDialog *dialog,
 
 // update interval picker callback - restrict value to UPDATE_INTERVAL_MIN - UPDATE_INTERVAL_MAX
 static void
-update_picker_changed_value_cb (HildonPickerButton *picker, gpointer data)
+update_picker_changed_value_cb (HildonPickerButton *picker, gpointer data G_GNUC_UNUSED)
 {
 	gdouble number = 0;
 
@@ -1634,8 +1660,9 @@ ccchh_dooris_desktop_widget_finalize (GObject *object)
 		self->update_url = NULL;
 	}
 
-	g_static_mutex_free(&(self->button_lock));
 	g_static_mutex_free(&(self->fetch_lock));
+	g_static_mutex_free(&(self->button_lock));
+	g_static_mutex_free(&(self->expose_lock));
 
 /*
 	if (self->gvfs_handle) {
@@ -1720,27 +1747,29 @@ ccchh_dooris_desktop_widget_init (CCCHHDoorIsDesktopWidget *self)
 
 	// initialize state
 	// this combination ensures that update() sets an image on first call
-	// TODO: is there a better way?
 	self->current_state = DOORIS_FNORD;
 	self->last_state = -1;
 	self->last_known_state = DOORIS_FNORD;
 
-	self->last_change_ts = 0L;
-	self->last_report_ts = 0L;
-	self->too_old = FALSE;
-
-	self->desktop_visible = TRUE;
-
 	// set image to NULL, first expose event will then call update()
 	self->image_current = NULL;
 	self->current_alpha = IMAGE_ALPHA_NORMAL;
 
-	g_static_mutex_init(&(self->button_lock));
+	self->last_change_ts = 0L;
+	self->last_report_ts = 0L;
+	self->too_old = FALSE;
+
 	g_static_mutex_init(&(self->fetch_lock));
+	g_static_mutex_init(&(self->button_lock));
+	g_static_mutex_init(&(self->expose_lock));
 
 	self->gconf_client = gconf_client_get_default();
 
+	self->desktop_visible = TRUE;
+
 	self->force_banner = FALSE;
+	self->redraw_required = FALSE;
+	self->updating = FALSE;
 
 	self->disconnecting_iap = NULL;
 	self->connected_bearer = NULL;
