@@ -107,11 +107,13 @@ typedef struct {
 	gboolean too_old;
 	gboolean desktop_visible;
 	gboolean redraw_required;
+	gboolean updating;
 	gchar *disconnecting_iap;
 	gchar *connected_bearer;
 	gdouble current_alpha;
-	GStaticMutex button_lock;
 	GStaticMutex fetch_lock;
+	GStaticMutex button_lock;
+	GStaticMutex expose_lock;
 	GConfClient *gconf_client;
 
 	// door state information
